###
# #%L
# aiSSEMBLE::Test::MDA::Machine Learning::Machine Learning Training
# %%
# Copyright (C) 2021 Booz Allen
# %%
# This software package is licensed under the Booz Allen Public License. All Rights Reserved.
# #L%
###
"""
Base implementation of this pipeline.

GENERATED CODE - DO NOT MODIFY (add your customizations in Bootstrap).

Generated from: templates/pipeline-common/generic.base.py.vm
"""

from abc import ABC

class BootstrapBase(ABC):
    def __init__(self):
        pass
    
